import java . util . Vector;
import java . math . BigInteger;
import java . io . IOException;



class InstructionReader
{



	private final TokenReader reader;
	private boolean second_pass;
	private int instruction_counter;
	private IntArray label_values;



	InstructionReader (java . io . File file)
		throws java . io . FileNotFoundException
	{
		reader = new TokenReader (file);
		second_pass = false;
		instruction_counter = 0;
		label_values = new IntArray ();
	}



	void reset (boolean second_pass) throws IOException
	{
		reader . reset ();
		instruction_counter = 0;
		this . second_pass = second_pass;
	}



	private Token peek_token () throws IOException
	{
		return reader . peek_token ();
	}



	private Token read_token () throws IOException
	{
		return reader . read_token ();
	}



	Instruction read_instruction () throws IOException, CompileError
	{
		++ instruction_counter;
		Token token = peek_token ();
		while (token . type == Token . LABEL)
		{
			parse_label_definition ();
			token = peek_token ();
		}
		switch (token . type)
		{
		case Token . Z:
			return parse_Z_instruction ();
		case Token . S:
			return parse_S_instruction ();
		case Token . T:
			return parse_T_instruction ();
		case Token . J:
			return parse_J_instruction ();
		case Token . EOF:
			return null;
		default:
			throw new CompileError (CompileError . LABEL_OR_INSTRUCTION_EXPECTED, token);
		}
	}



	private void parse_label_definition () throws IOException, CompileError
	{
		// Get label identifier:
		Token label = read_token ();

		// Make sure colon follows:
		Token colon = read_token ();
		if (colon . type != Token . COLON)
			throw new CompileError (CompileError . COLON_EXPECTED_AFTER_LABEL, colon);

		// Try to resolve label value:
		int index = ((LabelToken) label) . index;
		int value = label_values . get (index);
		if (value == 0)
		{
			label_values . set (index, instruction_counter);
		}
		else
			if (value != instruction_counter)
				throw new CompileError (CompileError . DUPLICATE_LABEL, label);
	}



	private void parse_open_paren () throws IOException, CompileError
	{
		Token token = read_token ();
		if (token . type != Token . OPEN_PAREN)
			throw new CompileError (CompileError . OPEN_PAREN_EXPECTED, token);
	}



	private void parse_close_paren () throws IOException, CompileError
	{
		Token token = read_token ();
		if (token . type != Token . CLOSE_PAREN)
			throw new CompileError (CompileError . CLOSE_PAREN_EXPECTED, token);
	}



	private void parse_comma () throws IOException, CompileError
	{
		Token token = read_token ();
		if (token . type != Token . COMMA)
			throw new CompileError (CompileError . COMMA_EXPECTED, token);
	}



	private BigInteger parse_register_number () throws IOException, CompileError
	{
		Token token = read_token ();
		if (token . type != Token . INTEGER)
			throw new CompileError (CompileError . POSITIVE_INTEGER_EXPECTED, token);
		BigInteger integer = ((IntegerToken) token) . integer;
		if (integer . compareTo (BigInteger . valueOf (0)) <= 0)
			throw new CompileError (CompileError . POSITIVE_INTEGER_EXPECTED, token);
		return integer;
	}



	private BigInteger parse_jump_target () throws IOException, CompileError
	{
		Token token = read_token ();
		BigInteger integer;
		switch (token . type)
		{
		case Token . INTEGER:
			integer = ((IntegerToken) token) . integer;
			if (integer . compareTo (BigInteger . valueOf (0)) <= 0)
				throw new CompileError (CompileError . POSITIVE_INTEGER_EXPECTED, token);
			break;
		case Token . LABEL:
			int target = label_values . get (((LabelToken) token) . index);
			if (target < 1 && second_pass)
				throw new CompileError (CompileError . UNDEFINED_LABEL, token);
			integer = BigInteger . valueOf (target);
			break;
		default:
			throw new CompileError (CompileError . POSITIVE_INTEGER_EXPECTED, token);
		}
		return integer;
	}



	private Z_Instruction parse_Z_instruction () throws IOException, CompileError
	{
		if (read_token () . type != Token . Z)
			throw new RuntimeException
			(
				"assertion failed: the next token was not a Z token"
			);
		parse_open_paren ();
		BigInteger dest = parse_register_number ();
		parse_close_paren ();
		return new Z_Instruction (dest, true);
	}



	private S_Instruction parse_S_instruction () throws IOException, CompileError
	{
		if (read_token () . type != Token . S)
			throw new RuntimeException
			(
				"assertion failed: the next token was not a S token"
			);
		parse_open_paren ();
		BigInteger dest = parse_register_number ();
		parse_close_paren ();
		return new S_Instruction (dest, true);
	}



	private T_Instruction parse_T_instruction () throws IOException, CompileError
	{
		if (read_token () . type != Token . T)
			throw new RuntimeException
			(
				"assertion failed: the next token was not a T token"
			);
		parse_open_paren ();
		BigInteger src = parse_register_number ();
		parse_comma ();
		BigInteger dest = parse_register_number ();
		parse_close_paren ();
		return new T_Instruction (src, dest);
	}



	private J_Instruction parse_J_instruction () throws IOException, CompileError
	{
		if (read_token () . type != Token . J)
			throw new RuntimeException
			(
				"assertion failed: the next token was not a J token"
			);
		parse_open_paren ();
		BigInteger src1 = parse_register_number ();
		parse_comma ();
		BigInteger src2 = parse_register_number ();
		parse_comma ();
		BigInteger target = parse_jump_target ();
		parse_close_paren ();
		return new J_Instruction (src1, src2, target);
	}



}
