﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace com.scrigner.RAMModel
{
	// ==============================================================================================================
	// Machine 
	// ==============================================================================================================
	// The Machine implements an Unlimited Register Machine with the following design choices.
 	// The machine defines a Tape in its state. This means that the building the machine, builds its tape also.
	// The Program is NOT part of the Machine, and it is submitted as an input to the Execute method.
	// The design choice of the program counter is tricky. 
	// The most natural design would define the PC as a mutable state of the machine. Apart from the mutability, the PC shou
	// .....
	// Initial value can be done in two ways : ......
	// ==============================================================================================================
	public class CMachine : IMachine
	{
		private const int MAX_STEPS_DEFAULT = 1000;
		private readonly ITape zTape;						// The Machine "almost-unlimited" Tape
		//private readonly IProgram zPProgram;
		//private readonly uint ziNextStatementIndex;		// Program Counter NO PERCHE'
		private readonly int ziMaxAllowedSteps;

		//private int ziNextStatementIndex;		// this is mutable (which all the problems it conveys) because the statement needs to update it
												// in model 1. A better model would make the statement return the next statement index

		public event EventHandler<CMachineStateEventArgs> NotifyProcessingState;	// The machine will notify its internal state (the tape and the program counter) at each processing iteration
		// ==============================================================================================================
		// CTORs
		// ==============================================================================================================
		public CMachine() : this (MAX_STEPS_DEFAULT) {}
		public CMachine(int iMaxAllowedSteps) : this (iMaxAllowedSteps, new long[]{}) {}
		public CMachine(IEnumerable<long> initialValues) : this (MAX_STEPS_DEFAULT, initialValues) {}
		public CMachine(int iMaxAllowedSteps, IEnumerable<long> initialValues)
		{
			CContract.Assert(iMaxAllowedSteps > 0, "iMaxAllowedSteps > 0");
			CContract.AssertIsNotNull("initialValues", initialValues);

			ziMaxAllowedSteps = iMaxAllowedSteps;
			zTape = new CTape(initialValues);
		}
		public ITape Tape	{get{return zTape;}}
		// ==============================================================================================================
		// Execute : The Machine executes a Program
		// ==============================================================================================================
		public long Execute(IProgram program)
		{
			CContract.AssertIsNotNull("program", program);

			int iStepCount = 0;					// We count the number of steps, in order to test the "quasi-loop" condition
			int iNextStatementIndex = 1;		// This is the Program Counter

			// We notify the initial state
			this.OnNotifyProcessingState(
					new CMachineStateEventArgs(
							new CMachineState(iStepCount, iNextStatementIndex, this, program)
							));

			// We iterate if we have another statement to process ...
			while (program.IsStatementDefinedAt(iNextStatementIndex))
			{	// On each cycle, we increment the step counter ...
				iStepCount++;
				// ... because if we reach the maximum count ...
				if (iStepCount > ziMaxAllowedSteps)
				{	// ... we treat it as a loop
					throw new CLoopDetectedException();
				}
				else
				{	// We ask the program to fetch the next statement to be executed ...
					IStatement nextStatement = program.GetStatement(iNextStatementIndex);
					// ... we delegate the execution to the statement and we get back the result ...
					IStatementResult result = nextStatement.Execute(this);

					// ... if the statement overrode the NextInstruction pointer ....
					if (result.HasNextStatementOverride())
					{	// ... we use it to update the program countet ...
						iNextStatementIndex = result.GetNextStatementIndex();
					}
					else
					{	// ... otherwise we simply go to the next statement (which is the default behaviour)
						iNextStatementIndex++;
					}

					// We notify the processing step ...
					this.OnNotifyProcessingState(
							new CMachineStateEventArgs(
									new CMachineState(iStepCount, iNextStatementIndex, this, program)
									));
				}
			}

			// We read the value at the FIRST position of the tape ...
			long returnValue = zTape.GetValue(1);
			// ... and we return it as a computation result ...
			return returnValue;
		}

		// ==============================================================================================================
		// Event Helpers - Event/Observer pattern
		// ==============================================================================================================
		protected virtual void OnNotifyProcessingState(CMachineStateEventArgs e)
		{
			if (NotifyProcessingState != null)
			{
				NotifyProcessingState.Invoke(this, e);
			}
		}
	}
}
