﻿using System;
using System.Collections.Generic;
using System.Text;

namespace com.scrigner.RAMModel.Parsing
{
	// ================================================================================
	// The Program Builder builds a IProgram from a string
	// ================================================================================
	// Since the project point is NOT the parser, this is implemented with a quick and dirty approach
	// ================================================================================
	public class CProgramParser
	{
		// ================================================================================
		// CTORs
		// ================================================================================
		public CProgramParser()
		{}
		// ================================================================================
		// Parse
		// ================================================================================
		public IProgram Parse(string source)
		{
			CContract.AssertIsNotNull("source", source);

			List<IStatement> statements = new List<IStatement>();

			foreach (string sStatement in this.EnumerateStatements(source))
			{
				IStatement statement = null;
				if (sStatement[0] == 'z')		{statement = new CStatementParserZero().Parse(sStatement);}
				else if (sStatement[0] == 's')	{statement = new CStatementParserAddOne().Parse(sStatement);}
				else if (sStatement[0] == 't')	{statement = new CStatementParserTransfer().Parse(sStatement);}
				else if (sStatement[0] == 'c')	{statement = new CStatementParserCondition().Parse(sStatement);}
				else
				{
					throw new CParsingException("Parsing Error on TextStatement=[" + sStatement + "]");
				}
				// ... we add the parsed statement to the list ....
				statements.Add(statement);
			}
			// ... we build a program
			IProgram program = new CProgram(statements);
			// ... we return it 
			return program;
		}

		// Splits source in n statements, ignores SPACE, \r \n ';' 
		private IEnumerable<string> EnumerateStatements(string source)
		{
			StringBuilder sb = new StringBuilder();
			foreach (char c in source)
			{
				if (c == ' ' || c == '\r')
				{}
				else if (c == '\n' || c == ';')
				{
					yield return sb.ToString();
					sb.Clear();
				}
				else
				{
					sb.Append(c);
				}
			}
			if (sb.Length > 0)
			{
				yield return sb.ToString();
			}
		}

		// ================================================================================
		// Nested classes : Statement Parsers
		// ================================================================================
		private abstract class CAbstractStatementParser
		{
			private readonly char zcFunctionName;

			public CAbstractStatementParser(char cFunctionName)
			{
				CContract.AssertIsNotNull("cFunctionName", cFunctionName);

				zcFunctionName = cFunctionName;
			}

			public IStatement Parse(string source)
			{
				CContract.AssertIsNotEmpty("source", source);

				if (source[0] == zcFunctionName && source[1] == '(' && source[source.Length-1] == ')')
				{
					string sArgument = source.Substring(2, source.Length-3);

					IStatement statement = this.DoParse(sArgument);

					return statement;
				}
				else
				{
					throw new CParsingException("Cannot Parse source=[" + source + "] with StatementParser of Type=[" + this.GetType().FullName + "]");
				}
			}
			protected abstract IStatement DoParse(string sArgument);
		}

		private class CStatementParserZero : CAbstractStatementParser 
		{
			public CStatementParserZero() : base ('z')
			{}
			protected override IStatement DoParse(string sArgument)
			{
				try
				{
					int n = (int)uint.Parse(sArgument);

					return new CStatementZero(n);
				}
				catch (Exception ex)
				{
					throw new CParsingException("Cannot ParseZ Argument=[" + sArgument + "]");
				}
			}
		}

		private class CStatementParserAddOne : CAbstractStatementParser 
		{
			public CStatementParserAddOne() : base ('s')
			{}
			protected override IStatement DoParse(string sArgument)
			{
				try
				{
					int n = (int)uint.Parse(sArgument);

					return new CStatementAddOne(n);
				}
				catch (Exception ex)
				{
					throw new CParsingException("Cannot ParseS Argument=[" + sArgument + "]");
				}
			}
		}

		private class CStatementParserTransfer : CAbstractStatementParser 
		{
			public CStatementParserTransfer() : base ('t')
			{}
			protected override IStatement DoParse(string sArgument)
			{
				string[] args = sArgument.Split(',');
				if (args.Length == 2)
				{
					try
					{
						int m = (int)uint.Parse(args[0]);
						int n = (int)uint.Parse(args[1]);

						return new CStatementTransfer(m, n);
					}
					catch (Exception ex)
					{
						throw new CParsingException("Cannot ParseT Argument=[" + sArgument + "]");
					}
				}
				else
				{
					throw new CParsingException("Cannot ParseT Argument=[" + sArgument + "]");
				}
			}
		}

		private class CStatementParserCondition : CAbstractStatementParser 
		{
			public CStatementParserCondition() : base ('c')
			{}
			protected override IStatement DoParse(string sArgument)
			{
				string[] args = sArgument.Split(',');
				if (args.Length == 3)
				{
					try
					{
						int m = (int)uint.Parse(args[0]);
						int n = (int)uint.Parse(args[1]);
						int q = (int)uint.Parse(args[2]);

						return new CStatementCondition(m, n, q);
					}
					catch (Exception ex)
					{
						throw new CParsingException("Cannot ParseC Argument=[" + sArgument + "]");
					}
				}
				else
				{
					throw new CParsingException("Cannot ParseC Argument=[" + sArgument + "]");
				}
			}
		}

	}
}
